import java.util.Vector;
import java.util.Hashtable;
import java.util.StringTokenizer;
import java.io.*;

/** DatalogParser - simple recursive descent parser for CSE 544/Tukwila extended datalog.  
 *  Output is a logical query plan consisting of children of the Operator class.  
 *
 *  Note that attributes in each of the operators are mapped from a symbol table.  Each 
 *  attribute is specified with an EDB prefix, i.e. it is specified as edb.attr 
 *  instead of simply attr.  This is to help you map the symbols correctly during
 *  conflicts.
 *
 * Basic input:  extended datalog query, single rule, e.g.:
 *
 * q(X) :- edb1(X), edb2(X, Y), edb3(X, Y, Z), {edb4, edb5}(Y, Z), X = 5
 *
 * (See http://www.cs.washington.edu/education/courses/cse544/CurrentQtr/collectors.html)
 *
 * Generates collector, project, select, join, (wrapper) scan
 *
 *
 * Thanks to Rachel Pottinger and Mark Berryman for providing me with code that gave me ideas
 * for attacking the problem!
 *
 * @author Zack Ives
 * @version 1.0
 */
 
class DatalogParser {
    Hashtable _symbolTable;
    Hashtable _tables;
    static String _dummy = "";
    String _idb;
    
    // Constructor - do nothing
    public DatalogParser() {
        _symbolTable = new Hashtable();
        _tables = new Hashtable();
    }

    // Read and return the next token, skipping any spaces
    public String getNextToken(StringTokenizer t) {
        String token = t.nextToken();
        
        // Skip any spaces
        while (token.equals(" ")) {
            if (!t.hasMoreTokens())
                return null;
            else    
                token = t.nextToken();
        }
        return token;    
    }

    // Parse the head of the rule, returning a Project operator and mapping the appropriate
    // symbols into the symbol table
    Operator parseHead(StringTokenizer t) {
        Vector currentAttrib = new Vector();
        
        // Error if no tokens
        if (!t.hasMoreTokens())
            return null;
                
        String output = getNextToken(t);
        
        if (!t.hasMoreTokens() || !(getNextToken(t).equals("(")))
            return null;
        
        String token = "";
        
        // Read tokens up to the close-paren
        token = getNextToken(t);
        while (t.hasMoreTokens() && !(token.equals(")"))) {
            if (!token.equals(",")) {
                currentAttrib.addElement(token);
                
                // If the attribute isn't in the symbol table, add it in
                if (!_symbolTable.containsKey(token))
                    _symbolTable.put(token, _dummy);
            }
            token = getNextToken(t);
        }
        if (token.equals(")")) {
            // Return project operator if we've got two more tokens equal to :-
            if ( t.hasMoreTokens() && getNextToken(t).equals(":")  && 
                 t.hasMoreTokens() && getNextToken(t).equals("-")      )
                return new Project(null, currentAttrib, _symbolTable);
        }
        return null;
    }
    
    // Parse the scan operation
    Operator parseScan(String edb, StringTokenizer t) {
        Vector currentAttrib = new Vector();
        String token;
        
        // First check that the table is defined
        if (!_tables.containsKey(edb)) {
            System.err.println("Unable to find table " + edb + " in our list.");
            return null;
        }
        
        Table table = (Table)_tables.get(edb);

        // Read tokens up to the close-paren
        token = getNextToken(t);
        int index = 0;
        while (t.hasMoreTokens() && !(token.equals(")"))) {
            if (!token.equals(",")) {
                currentAttrib.addElement(token);
                    
                // If the attribute isn't in the symbol table, add it in
                if (!_symbolTable.containsKey(token) || 
                    _symbolTable.get(token) == _dummy) {

                    // Look up the real attribute name in the symbol table    
                    String attrName = table.getAttribute(index++);
                    
                    if (index > table.size()) {
                        System.err.println("Error!  Datalog EDB " + edb + " has incorrect number of variables!");
                    
                        return null;
                    }

                    if (attrName == null) {
                        System.err.println("Symbol " + token + " not found.");
                        return null;
                    }

                    // Save the mapping into the symbol table; it will get replaced later
                    // as needed by various operators
                    _symbolTable.put(token, edb + "." + attrName);
                }
            }
            token = getNextToken(t);
        }
        if (table.size() != currentAttrib.size()) {
            System.err.println("Error!  Datalog EDB " + edb + " has incorrect number of variables!");
        
            return null;
        }
        if (token.equals(")"))
            return new Scan(edb, currentAttrib, table, _symbolTable);
        return null;
    }
    
    // Parse the collector operator
    Operator parseCollector(StringTokenizer t) {
        Vector alternateSources = new Vector();
        Vector currentAttrib = new Vector();
        Vector children = new Vector();
        String token;
        
        if (!t.hasMoreTokens())
            return null;

        // Parse each of the alternate tables in the collector
        token = getNextToken(t);
        while (t.hasMoreTokens() && !(token.equals("}"))) {
            if (!token.equals(",")) {
                alternateSources.addElement(token);
            }
            token = getNextToken(t);
        }
        String edb = null;
        Table table = null;
        
        for (int sourceNo = 0; sourceNo < alternateSources.size(); sourceNo++) {
            edb = (String)alternateSources.elementAt(sourceNo);
            
            // First check that the table is defined
            if (!_tables.containsKey(edb)) {
                System.err.println("Unable to find table " + edb + " in our list.");
                return null;
            }
        }
        
        // Go back to the first table
        edb = (String)alternateSources.elementAt(0);
            
        // Get its attribute mappings
        table = (Table)_tables.get(edb);

        // Read tokens up to the close-paren
        token = getNextToken(t);
        if (!token.equals("("))
            return null;
            
        token = getNextToken(t);
        int index = 0;
        while (t.hasMoreTokens() && !(token.equals(")"))) {
            if (!token.equals(",")) {
                currentAttrib.addElement(token);
                        
                // If the attribute isn't in the symbol table, add it in
                // This means that, in the cases of joins, the leftmost table "wins"
                // in that its entry goes into the symbol table
                if (!_symbolTable.containsKey(token) || 
                    _symbolTable.get(token) == _dummy) {

                    // Look up the real attribute name in the symbol table    
                    String attrName = table.getAttribute(index++);

                    if (index > table.size()) {
                        System.err.println("Error!  Datalog EDB " + edb + " has incorrect number of variables!");
                    
                        return null;
                    }
                        
                    if (attrName == null) {
                        System.err.println("Symbol " + token + " not found.");
                        return null;
                    }

                    // Save the mapping into the symbol table
                    _symbolTable.put(token, edb + "." + attrName);
                }
            }
            token = getNextToken(t);
        }
        // Add a bunch of children with identical attribute sets
        if (token.equals(")"))
            for (int sourceNo = 0; sourceNo < alternateSources.size(); sourceNo++) {
                edb = (String)alternateSources.elementAt(sourceNo);
                table = (Table)_tables.get(edb);

                if (table.size() != currentAttrib.size()) {
                    System.err.println("Error!  Datalog EDB " + edb + " has incorrect number of variables!");
                
                    return null;
                }

                children.addElement(new 
                    Scan((String)alternateSources.elementAt(sourceNo), currentAttrib, 
                    table, _symbolTable));
            }        
        return new Collector(children, _symbolTable);
    }

    // Parse a selection statement, which is simply a built-in atom.  Note that this
    // operator is unlike the collector and the scan in that it actually has a child
    Operator parseSelect(String var, String op, StringTokenizer t, 
        Operator curRoot) {
            
        if (op.equals("=")) {
        } else if (op.equals("<")) {
        } else if (op.equals(">")) {
        } else if (op.equals(">=")) {
        } else if (op.equals("<=")) {
        } else if (op.equals("<>") || op.equals("!=")) {
        } else {
            System.err.println("Illegal built-in atom " + op + " for " + var + "!");
            return null;
        }
        
        // Make sure the variable has been defined before we select on it
        if (!_symbolTable.containsKey(var) || _symbolTable.get(var) == _dummy) {
            System.err.println("Select on undefined variable " + var + "!");
            return null;
        }
        
        String value = getNextToken(t);
        
        return new Select(curRoot, var, op, value, _symbolTable);
    }
    
    // Parses an atom (either built-in or an EDB)
    Operator parseAtom(StringTokenizer t, Operator curRoot) {
        String var = getNextToken(t);
        Operator curNode = null;
        
        if (!t.hasMoreTokens())
            return null;
            
        // If we've got a collector, it's a leaf
        if (var.equals("{"))
            curNode = parseCollector(t);
        else {
            String token = getNextToken(t);
            
            // Determine whether we have a scan or something else:
            if (token.equals("(")) {
                // Scan is a leaf
                curNode = parseScan(var, t);
            } else {
                // Select needs to know the current root node
                curNode = parseSelect(var, token, t, curRoot);
            }
        }

        // Consume a trailing comma if it exists
        if ( !t.hasMoreTokens() || getNextToken(t).equals(","))
            return curNode;
        else {
            System.err.println("Unable to parse atom for " + var);
            return null;
        }
    }
    
    // Parses the body of the operator tree, building a join tree in left-to-right
    // fashion (left linearly) and optionally adding a select.  Returns
    // the root of the tree.
    private Operator parseBody(StringTokenizer t) {
        Operator curRoot = null;
        Operator curNode;
        while (t.hasMoreTokens()) {
            curNode = parseAtom(t, curRoot);
            
            // If the node was null, we had an error - abort
            if (curNode == null)
                return null;
            
            // If this is the first atom, it's our entire tree.  If it's a select,
            // it was built on the old root, so we still need to replace our tree.
            if (curRoot == null || (curNode instanceof Select))
                curRoot = curNode;
                
            // Otherwise, join as part of a left-linear tree
            else
                // Needs to be fixed to add join conditions
                curRoot = new Join(curRoot, curNode, _symbolTable);
        }
        return curRoot;
    }
    
    // Parse a query and return the operator tree
    public Operator parse(String query) {
        _symbolTable = new Hashtable();
        Operator current;
        Vector currentAttrib = new Vector();
        Operator root;
        Vector rootAttrib = new Vector();
        StringTokenizer t = new StringTokenizer(query, ":-, (){}", true);
        
        root = parseHead(t);
        current = parseBody(t);
        
        if (current == null)
            return null;
        
        if (root != null) {
            if (((Project)root).setChild(current) == false)
                return null;
        }
        
        return root;
    }
    
//////////////////////////////// Misc functions here - you'll replace these!!!!
    
    // Sets up the EDB table definitions in a hash table so we can look up actual
    // schema info on tables
    public void addTables() {
        Table t = new Table("r1");
        t.addAttribute("alpha");
        t.addAttribute("beta");
        _tables.put(t.name(), t);
        
        t = new Table("r2");
        t.addAttribute("gamma");
        t.addAttribute("delta");
        _tables.put(t.name(), t);

        t = new Table("r3");
        t.addAttribute("epsilon");
        _tables.put(t.name(), t);

        t = new Table("r4");
        t.addAttribute("epsilon");
        _tables.put(t.name(), t);
        
        t = new Table("r5");
        t.addAttribute("epsilon");
        t.addAttribute("phi");
        _tables.put(t.name(), t);
    }
    
    // Simple test of query parsing - parse query and print it
    static void runQuery(DatalogParser d, String query) {
        Operator result;
        
        System.out.println(query);
        result = d.parse(query);
        if (result != null)
            System.out.println(result.toString());
        else
            System.err.println("Error parsing " + query);
    }
    
    //////////
    // Mainline - input a query and parse it
    static public void main(String args[]) {
        String query;
        
        DatalogParser d = new DatalogParser();
        DataInputStream i = new DataInputStream(System.in);
 
        // This method specifies metadata for all known tables
        // It should essentially be a subset of the data source catalog
        d.addTables();

        // A few sample queries:
        System.out.println("Sample queries:");
        runQuery(d, "q(X,Y) :- r1(Y, Z), r2(Z,X), {r3,r4}(Y), X = 5");
        runQuery(d, "q(X,Y) :- {r1,r5}(X, Y), Y = 3, r2(Z,Y), {r3,r4}(X)");
        runQuery(d, "q(X,Y) :- r1(Z,Q), Q = 5, r2(X,Z), r3(Y)");
        
        try {
            System.out.println("\nPlease enter a query, blank to quit: ");
            query = i.readLine();
            while (!query.equals("")) {
                runQuery(d, query);
                System.out.println("\nPlease enter a query, blank to quit: ");
                query = i.readLine();
            }
        } catch (java.io.IOException e) {
        }
    }
}